<?php

namespace ePHPic\View;

use ePHPic\View\Attributes\AccessKeyAttribute;
use ePHPic\View\Attributes\ClassAttribute;
use ePHPic\View\Attributes\ComponentCustomAttribute;
use ePHPic\View\Attributes\ContentEditableAttribute;
use ePHPic\View\Attributes\DirectionAttribute;
use ePHPic\View\Attributes\DraggableAttribute;
use ePHPic\View\Attributes\EnterKeyHintAttributes;
use ePHPic\View\Attributes\HiddenAttribute;
use ePHPic\View\Attributes\IdentifierAttribute;
use ePHPic\View\Attributes\InertAttribute;
use ePHPic\View\Attributes\InputModeAttribute;
use ePHPic\View\Attributes\LanguageAttribute;
use ePHPic\View\Attributes\PopoverAttribute;
use ePHPic\View\Attributes\SpellCheckAttribute;
use ePHPic\View\Attributes\StyleAttribute;
use ePHPic\View\Attributes\TabIndexAttribute;
use ePHPic\View\Attributes\TitleAttribute;
use ePHPic\View\Attributes\TranslateAttribute;
use ePHPic\Utilities\GUID;
use Exception;
use ReflectionClass;
use ReflectionNamedType;
use ReflectionProperty;

abstract class Tag implements RenderableInterface
{
    protected string $GUID;

    public AccessKeyAttribute $accessKey;
    public ClassAttribute $class;
    public ContentEditableAttribute $contenteditable;
    public DirectionAttribute $dir;
    public DraggableAttribute $draggable;
    public EnterKeyHintAttributes $enterkeyhint;
    public HiddenAttribute $hidden;
    public IdentifierAttribute $id;
    public InertAttribute $inert;
    public InputModeAttribute $inputmode;
    public LanguageAttribute $lang;
    public PopoverAttribute $popover;
    public SpellCheckAttribute $spellcheck;
    public StyleAttribute $style;
    public TabIndexAttribute $tabindex;
    public TitleAttribute $title;
    public TranslateAttribute $translate;
    public ComponentCustomAttribute $component;
    
    protected array $children = [];
    protected string $innerHTML;

    public function __construct(
        string $innerHTML = "",
        array $attributes = [],
        array $children = []
    )
    {
        $this->GUID = GUID::generate();
        $this->children = $children;
        $this->innerHTML = $innerHTML;
        $reflection = new \ReflectionClass($this);
        foreach ($reflection->getProperties(ReflectionProperty::IS_PUBLIC) as $property)
        {
            
            $key = $property->getName();
            if (array_key_exists($key, $attributes))
            {
                $this->$key = $attributes[$key];
            }
            else
            {
                $type = $property->getType();
                if ($type instanceof ReflectionNamedType && !$type->isBuiltin())
                {
                    $className = $type->getName();
                    $this->$key = $className::fromValue(null);
                }
                else
                {
                    $this->$key = null;
                }
            }
        }
    }
    public static function make(
        string $innerHTML = "",
        array $attributes = [],
        array $children = []
    ): static
    {
        return new static($innerHTML, $attributes, $children);
    }
    public function render(int $indent = 0, bool $minification = false): string
    {
        $tabs = str_repeat("\t", $indent);
        $newLine = "\n";
        if(true === $minification)
        {
            $tabs = "";
            $newLine = "";
        }
        $tag = $this->getTag();
        
        $output = "$tabs<$tag";
        
        $reflection = new ReflectionClass($this);
        foreach ($reflection->getProperties(ReflectionProperty::IS_PUBLIC) as $property) {
            $value = $property->getValue($this);
            if ($value !== null) {
                if(is_object($value) && method_exists($value, "render"))
                {
                    $attribute = $value->render();
                }
                if(false === empty($attribute))
                {
                    $output .= " " . $value->render();
                }
            }
        }
        
        if ($this->isEmptyTag())
        {
            $output .= " />$newLine";
        }
        else
        {
            $output .= ">";

            if(0 < count($this->children))
            {
                $output .= $newLine;
            }
            
            $output .= $this->innerHTML ?? '';
            
            if(0 < count($this->children) && !empty($this->innerHTML))
            {
                $output .= $newLine;
            }

            foreach ($this->children as $child)
            {
                if ($child instanceof RenderableInterface) {
                    $output .= $child->render($indent+1);
                }
            }
            if(0 < count($this->children))
            {
                $output .= $tabs;
            }
            $output .= "</$tag>$newLine";
        }

        return $output;
    }
    public function AddChildren(string $key, Tag $children): void
    {
        $this->children[$key] = $children;
    }
    public function getChildern(string $key): ?Tag
    {
        return $this->children[$key] ?? null;
    }
    public function removeChildren(string $key): void
    {
        unset($this->children[$key]);
    }
    public function setInnerHTML(string $html): void
    {
        $this->innerHTML = $html;
    }
    public function getInnerHTML(): ?string
    {
        return $this->innerHTML;
    }
    public function appendHTML(string $html): void
    {
        $this->innerHTML .= $html;
    }
    public function prependHTML(string $html): void
    {
        $this->innerHTML = $html . $this->innerHTML;
    }
    public function setAttribute($attributeName, $value)
    {
        $reflection = new ReflectionClass($this);
        if ($reflection->hasProperty($attributeName))
        {
            $property = $reflection->getProperty($attributeName);
            if ($property->isPublic())
            {
                $type = $property->getType();
                if ($type instanceof ReflectionNamedType && !$type->isBuiltin())
                {
                    $className = $type->getName();
                    $this->$attributeName = $className::fromValue($value);
                }
            }
        }
    }
    public function getAttribute(string $attributeName)
    {
        $reflection = new ReflectionClass($this);
        if ($reflection->hasProperty($attributeName))
        {
            $property = $reflection->getProperty($attributeName);
            if ($property->isPublic())
            {
                $value = $property->getValue($this);
                $type = $property->getType();
                if ($type instanceof ReflectionNamedType && !$type->isBuiltin())
                {
                    return $value->value();
                }
                else
                {
                    return $value;
                }
            }
            else
            {
                throw new Exception("Property $attributeName is not public.");
            }
        } else {
            throw new Exception("Property $attributeName does not exist.");
        }
    }
    public function getGUID(): string
    {
        return $this->GUID;
    }
    abstract protected function getTag(): string;
    abstract protected function isEmptyTag(): bool;
}
